/* ttt - a tic tac toe game for any ASCII capable console
 * Copyright (C) 2006 by Erik Auerswald <auerswal@unix-ag.uni-kl.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define VERSION "1.0.1"

#define BUFFERSIZE 80				/* size of input buffer */

/* global variables to avoid passing board[] and stone[] to every function */
/* 9 fields, space represents empty */
char board[9] = { ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ' };
/* player 0 uses 'x', player 1 'o' */
char stone[2] = { 'x', 'o' };

/* print board to stdout */
static void print_board(void) {
  int i;
#ifdef PRINT_SMALL
  
  puts("\n|1|2|3|\n+-+-+-+-");
  for(i=1; i<10; i++) {
    printf("|%c", board[i-1]);
    if(i%3 == 0) {				/* 3 fields per row */
      printf("|%d\n+-+-+-+-\n", i/3);
    }
  }
  puts("");
#else
  puts("\n      1   2   3\n");
  for(i=0; i<3; i++) {
    int j;
    printf("  %d  ",i+1);
    for(j=0;j<3;j++) {
      printf(" %c ",board[i*3+j]);
      if(j<2) {
	printf("|");
      }
    }
    if(i<2) {
      printf("\n     ---+---+---\n");
    }
  }
  printf("\n\n");
#endif
}

/* check for end of game conditions */
static int wincondition(void) {
  int i, player;

  /* search for 3 stones in a row */
  for(player=0; player<2; player++) {
    /* check rows */
    for(i=0; i<3; i++) {
      if((board[i*3+0] == board[i*3+1]) && (board[i*3+1] == board[i*3+2]) &&
	 (board[i*3+0] == stone[player])) {
	return player+1;
      }
    }
    /* check columns */
    for(i=0; i<3; i++) {
      if((board[i] == board[i+3]) && (board[i+3] == board[i+6]) &&
	 (board[i] == stone[player])) {
	return player+1;
      }
    }
    /* check diagonals */
    if(((board[0+0] == board[1+3]) && (board[1+3] == board[2+6]) &&
       (board[0+0] == stone[player])) ||
      ((board[2+0] == board[1+3]) && (board[1+3] == board[0+6]) &&
       (board[2+0] == stone[player]))) {
      return player+1;
    }
  }
  /* check for draw */
  /* if every field is used the game is a draw */
  for(i=0;i<9;i++) {
    if(board[i] == ' ') {		/* an empty field => game continues */
      return 0;
    }
  }
  return 3;				/* draw game */
}

/* place a stone on the board */
static int place_stone(int x, int y, int player) {
  if(board[3*y+x] != ' ') {
    fprintf(stderr, "Error: Placing stone %c on non empty field %d,%d\n",
            stone[player], x, y);
  }
  board[3*y+x] = stone[player];		/* 0 <= x, y <= 2 */
  return wincondition();		/* game over? */
}

/* this strategy uses only the current situation on the board */
static int strategy(int player) {	/* turn is for player player */
  int i;
  int opponent = (player+1) % 2;	/* number of opponent */

  printf("Player %d (%c) is controlled by the CPU...",player+1,stone[player]);
  /** find winning move **/
#ifdef DEBUG
puts("\ntrying heuristic 'looking for winning move'...");
#endif
  /* check rows */
  for(i=0; i<3; i++) {
    if((board[3*i+0] == stone[player]) && (board[3*i+1] == board[3*i+0]) &&
       (board[3*i+2] == ' ')) {
      puts(" :-D");
      return place_stone(2, i, player);
    } else if((board[3*i+0] == stone[player]) &&
              (board[3*i+2] == board[3*i+0]) && (board[3*i+1] == ' ')) {
      puts(" :-D");
      return place_stone(1, i, player);
    } else if((board[3*i+1] == stone[player]) &&
	      (board[3*i+2] == board[3*i+1]) && (board[3*i+0] == ' ')) {
      puts(" :-D");
      return place_stone(0, i, player);
    }
  }
  /* check columns */
  for(i=0; i<3; i++) {
    if((board[0+i] == stone[player]) && (board[3+i] == board[0+i]) &&
       (board[6+i] == ' ')) {
      puts(" :-D");
      return place_stone(i, 2, player);
    } else if((board[0+i] == stone[player]) && (board[6+i] == board[0+i]) &&
	      (board[3+i] == ' ')) {
      puts(" :-D");
      return place_stone(i, 1, player);
    } else if((board[3+i] == stone[player]) && (board[6+i] == board[3+i]) &&
	(board[0+i] == ' ')) {
      puts(" :-D");
      return place_stone(i, 0, player);
    }
  }
  /* check digonals */
  if((board[0]==stone[player]) && (board[4]==board[0]) && (board[8]==' ')) {
    return place_stone(2, 2, player);
  } else if((((board[0]==stone[player]) && (board[8]==board[0])) ||
             ((board[2]==stone[player]) && (board[6]==board[2]))) &&
            (board[4]==' ')) {
    puts(" :-D");
    return place_stone(1, 1, player);
  } else if((board[4]==stone[player]) && (board[8]==board[4]) &&
            (board[0]==' ')) {
    puts(" :-D");
    return place_stone(0, 0, player);
  } else if((board[2]==stone[player]) && (board[4]==board[2]) &&
            (board[6]==' ')) {
    puts(" :-D");
    return place_stone(0, 2, player);
  } else if((board[4]==stone[player]) && (board[6]==board[4]) &&
            (board[2]==' ')) {
    puts(" :-D");
    return place_stone(2, 0, player);
  }
  /** block winning move of opponent **/
#ifdef DEBUG
puts("trying heuristic 'looking for winning move of opponent to block'...");
#endif
  /* check rows */
  for(i=0; i<3; i++) {			/* i is y coordinate */
    if((board[3*i+0] == stone[opponent]) && (board[3*i+1] == board[3*i+0]) &&
       (board[3*i+2] == ' ')) {
      puts(" :-P");
      return place_stone(2, i, player);
    } else if((board[3*i+0] == stone[opponent]) &&
              (board[3*i+2] == board[3*i+0]) && (board[3*i+1] == ' ')) {
      puts(" :-P");
      return place_stone(1, i, player);
    } else if((board[3*i+1] == stone[opponent]) &&
	      (board[3*i+2] == board[3*i+1]) && (board[3*i+0] == ' ')) {
      puts(" :-P");
      return place_stone(0, i, player);
    }
  }
  /* check columns */
  for(i=0; i<3; i++) {			/* i is x coordinate */
    if((board[0+i] == stone[opponent]) && (board[3+i] == board[0+i]) &&
       (board[6+i] == ' ')) {
      puts(" :-P");
      return place_stone(i, 2, player);
    } else if((board[0+i] == stone[opponent]) && (board[6+i] == board[0+i]) &&
	      (board[3+i] == ' ')) {
      puts(" :-P");
      return place_stone(i, 1, player);
    } else if((board[3+i] == stone[opponent]) && (board[6+i] == board[3+i]) &&
	(board[0+i] == ' ')) {
      puts(" :-P");
      return place_stone(i, 0, player);
    }
  }
  /* check digonals */
  if((board[0]==stone[opponent]) && (board[4]==board[0]) && (board[8]==' ')) {
    return place_stone(2, 2, player);
  } else if((((board[0]==stone[opponent]) && (board[8]==board[0])) ||
	     ((board[2]==stone[opponent]) && (board[6]==board[2]))) &&
            (board[4]==' ')) {
    puts(" :-P");
    return place_stone(1, 1, player);
  } else if((board[4]==stone[opponent]) && (board[8]==board[4]) &&
            (board[0]==' ')) {
    puts(" :-P");
    return place_stone(0, 0, player);
  } else if((board[2]==stone[opponent]) && (board[4]==board[2]) &&
            (board[6]==' ')) {
    puts(" :-P");
    return place_stone(0, 2, player);
  } else if((board[4]==stone[opponent]) && (board[6]==board[4]) &&
            (board[2]==' ')) {
    puts(" :-P");
    return place_stone(2, 0, player);
  }
  /** secure middle field if possible **/
#ifdef DEBUG
puts("trying heuristic 'securing middle position'...");
#endif
  if(board[4] == ' ') {			/* 4 == 3*1 + 1 */
    puts(" :-)");
    return place_stone(1, 1, player);
  }
  /** place in opposing corner **/
#ifdef DEBUG
puts("trying heuristic 'placing in opposite corner'...");
#endif
  /* this is intended for the third stone in the game */
  if(board[4] == stone[player]) {
    if((board[0] == ' ') && (board[8] == stone[opponent])) {
      puts(" :-/");
      return place_stone(0, 0, player);
    } else if((board[2] == ' ') && (board[6] == stone[opponent])) {
      puts(" :-/");
      return place_stone(2, 0, player);
    } else if((board[6] == ' ') && (board[2] == stone[opponent])) {
      puts(" :-/");
      return place_stone(0, 2, player);
    } else if((board[8] == ' ') && (board[0] == stone[opponent])) {
      puts(" :-/");
      return place_stone(2, 2, player);
    }
  }
  /** place in corner of empty side **/
#ifdef DEBUG
puts("trying heuristic 'placing in corner of empty side'...");
#endif
  /* this is intended for the second stone in the game */
  if(board[4] == stone[opponent]) {
    if((board[0] == ' ') && (((board[1] == ' ') && (board[2] == ' ')) ||
	                     ((board[3] == ' ') && (board[6] == ' ')))) {
      puts(" :-/");
      return place_stone(0, 0, player);
    } else if((board[8] == ' ') &&
              (((board[7] == ' ') && (board[6] == ' ')) ||
	       ((board[2] == ' ') && (board[5] == ' ')))) {
      puts(" :-/");
      return place_stone(2, 2, player);
    }
  }
  /** place in corner opposite of non empty side (weak heuristic!) **/
#ifdef DEBUG
puts("trying heuristic 'placing in corner opposite of non empty middle field'...");
#endif
  /* this is intended for the third stone in the game */
  if(board[4] == stone[player]) {
    if(board[1] == stone[opponent]) {
      if(board[6] == ' ') {
	puts(" :-/");
	return place_stone(0, 2, player);
      } else if(board[8] == ' ') {
	puts(" :-/");
	return place_stone(2, 2, player);
      }
    } else if(board[3] == stone[opponent]) {
      if(board[2] == ' ') {
	puts(" :-/");
	return place_stone(2, 0, player);
      } else if(board[8] == ' ') {
	puts(" :-/");
	return place_stone(2, 2, player);
      }
    } else if(board[5] == stone[opponent]) {
      if(board[0] == ' ') {
	puts(" :-/");
	return place_stone(0, 0, player);
      } else if(board[6] == ' ') {
	puts(" :-/");
	return place_stone(0, 2, player);
      }
    } else if(board[7] == stone[opponent]) {
      if(board[0] == ' ') {
	puts(" :-/");
	return place_stone(0, 0, player);
      } else if(board[2] == ' ') {
	puts(" :-/");
	return place_stone(2, 0, player);
      }
    }
  }
  /** place in first free corner **/
#ifdef DEBUG
puts("trying heuristic 'placing in first free corner'...");
#endif
  /* this is at least better than placing in the first free position */
  if(board[0] == ' ') {
    puts(" :-/");
    return place_stone(0, 0, player);
  } else if(board[2] == ' ') {
    puts(" :-/");
    return place_stone(2, 0, player);
  } else if(board[6] == ' ') {
    puts(" :-/");
    return place_stone(0, 2, player);
  } else if(board[8] == ' ') {
    puts(" :-/");
    return place_stone(2, 2, player);
  }
  /** place in middle of empty row/column (very weak heuristic) **/
#ifdef DEBUG
puts("trying heuristic 'placing in middle of empty side'...");
#endif
  /* this is intended for the second stone in the game */
  if(board[4] == stone[opponent]) {
    if((board[0] == ' ') && (board[1] == ' ') && (board[2] == ' ')) {
      puts(" :-(");
      return place_stone(1, 0, player);
    } else if((board[6] == ' ') && (board[7] == ' ') && (board[8] == ' ')) {
      puts(" :-(");
      return place_stone(1, 2, player);
    } else if((board[0] == ' ') && (board[3] == ' ') && (board[6] == ' ')) {
      puts(" :-(");
      return place_stone(0, 1, player);
    } else if((board[2] == ' ') && (board[5] == ' ') && (board[8] == ' ')) {
      puts(" :-(");
      return place_stone(2, 1, player);
    }
  }
  /** find the first free field to place stone (not even a real heuristic) **/
#ifdef DEBUG
puts("using desperate strategy 'finding first empty field to place stone'...");
#endif
  /* this is a fallback if none of the heuristics found a move */
  /* I don't even know if this can happen */
  /* if this part is used it might be a flaw in the strategy that allows
   * a win against the computer */
  for(i=0; i<9; i++) {
    if(board[i] == ' ') {
      puts(" :-(");
      return place_stone(i%3, i/3, player);
    }
  }
  /* this code must not be reached */
  return -1;
}

/* read turn of human player from stdin and place stone on board */
static int human(int player) {
  int x=-1,y=-1;			/* initialize to illegal values */
  char buffer[BUFFERSIZE];		/* read line into this */

  buffer[BUFFERSIZE-1] = '\0';
  while(((x<1) || (x>3) || (y<1) || (y>3)) || (board[3*(y-1)+(x-1)] != ' ')) {
    printf("Enter coordinates x y (between 1 and 3), player %d: ", player+1);
    if(fgets(buffer, BUFFERSIZE-1, stdin)) {
      if(sscanf(buffer, "%d %d", &x, &y) != 2) {
	x = y = -1;
      }
    }
  }
  x--; y--;				/* array starts at position 0 */
  return place_stone(x, y, player);
}

/* one turn */
static int turn(int gametype, int player) {
  switch(gametype) {
    case 1:				/* human vs. human */
      return human(player);
      break;
    case 2:				/* human vs. computer */
      if(player == 0) {
	return human(player);
      } else {
	return strategy(player);
      }
      break;
    case 3:				/* computer vs. human */
      if(player == 1) {
	return human(player);
      } else {
	return strategy(player);
      }
      break;
    case 4:				/* computer vs. computer */
      return strategy(player);
      break;
    default:				/* this is an error */
      fprintf(stderr, "Error: unknown gametype %d\n", gametype);
      exit(EXIT_FAILURE);
      break;
  }
  /* this code must not be reached */
  return -1;
}

/* print game result to stdout */
static void show_result(int gametype, int result) {
  char *w_type;				/* type of winning player */

  if(result == 3) {			/* draw */
    puts("The game is a draw.\n");
  } else {
    if(((gametype==2) && (result==2)) || ((gametype==3) && (result==1)) ||
	(gametype==4)) {		/* computer won */
      w_type = strdup("computer");
    } else {				/* human won */
      w_type = strdup("human");
    }
    printf("Player %d (%s) has won\n\n", result, w_type);
    free(w_type);
  }
}

/* what the f*** does this program do? */
static void print_rules(void) {
  puts("\nThe rules:\n----------");
  puts("Place 3 stones in a row to win (horizontally, vertically or diagonally).");
  printf("Player 1 uses %c, player 2 %c.\n", stone[0], stone[1]);
  puts("Enter x and y coordinates seperated by whitespace (range 1-3).");
  puts("Example: 2 2 [places stone in middle field]\n");
}

/* a simple menu */
static int menu(void) {
  int input=0;
  char buffer[BUFFERSIZE];

  buffer[BUFFERSIZE-1]='\0';
  /* show instructions */
  puts("Choose one of the following:\n");
  puts("  (1) human vs. human");
  puts("  (2) human vs. computer");
  puts("  (3) computer vs. human");
  puts("  (4) computer vs. computer");
  puts("\n  (8) show the rules");
  puts("  (9) quit the game\n");
  /* get valid input */
  while((input<1) || ((input>4) && (input<8)) || (input>9)) {
    printf("Your choice? ");
    if(fgets(buffer, BUFFERSIZE-1, stdin)) {
      if(sscanf(buffer, "%d", &input) != 1) {
	input = 0;
      }
    }
  }
  return input;
}

static void print_welcome(void) {
  puts("+--------------------------------------+");
  puts("| Welcome to tic-tac-toe version " VERSION " |");
  puts("+--------------------------------------+\n");
}

static void play_game(int gametype) {
  int player;				/* 0 <= player <= 1 */
  int gameover;				/* if game is over and result */

  switch(gametype) {
    case 1: puts("\nPlaying a game human vs. human:"); break;
    case 2: puts("\nPlaying a game human vs. computer:"); break;
    case 3: puts("\nPlaying a game computer vs. human:"); break;
    case 4: puts("\nPlaying a game computer vs. computer:"); break;
    default: fprintf(stderr, "Error: illegal gametype %d\n", gametype); break;
  }
  player = gameover = 0;
  /* show empty board with coordinate system */
  print_board();
  /* game loop */
  while(!gameover) {			/* stop at end of game */
    gameover = turn(gametype, player);  /* process one turn */
    player = (player+1) % 2;		/* alternate players */
    print_board();			/* show what happened */
  }
  /* print game result */
  show_result(gametype, gameover);
}

static void init_board(void) {
  int i;
  for(i=0; i<9; i++) {
    board[i] = ' ';
  }
}

int main(void) {
  int gametype = 0;			/* who the players are */

  print_welcome();
  /* play the game */
  while((gametype = menu()) != 9) {
    init_board();
    switch(gametype) {
      case 1:
      case 2:
      case 3:
      case 4:
	play_game(gametype); break;
      case 8:
	print_rules(); break;
      default:
	break;
    }
  }
  exit(EXIT_SUCCESS);
}
